import logging
import functools

import paramiko
from typeguard import typechecked

from typing import Union, Callable, Optional
from . import exceptions
from . import _sftp_client

from ._sftp_client import (  # noqa; pylint: disable=unused-import
    DEFAULT_WINDOW_SIZE,
    DEFAULT_MAX_PACKET_SIZE,
    MAX_PACKET_SIZE,
    MAX_WINDOW_SIZE,
)

logger = logging.getLogger(__name__)


@typechecked
def initialize(
    *,
    address: str,
    port: Union[int, str] = 22,
    username: Optional[str] = None,
    password: Optional[str] = None,
    private_key: Optional[str] = None,
    passphrase: Optional[str] = None,
    connection_timeout: int = 10,
    transfer_timeout: int = (10 * 60),
    window_size: Optional[int] = DEFAULT_WINDOW_SIZE,
    max_packet_size: Optional[int] = DEFAULT_MAX_PACKET_SIZE,
    verify_host_key: bool = False,
) -> Callable:
    """ Initialize an SFTP session and get a connection context manager.

    :param address: The hostname or IP address of the device to connect to.
    :param port: The listening SSH port (usually 22).
    :param username: Optional username for :ref:`index:Authentication`.
    :param password: Optional password for :ref:`index:Authentication`. If supplied, can be used for
        private key decryption if a key is supplied with no passphrase argument.
    :param private_key: Optional private key to use for key :ref:`index:Authentication`.
    :param passphrase: Used for decrypting the private key (see :ref:`index:Authentication`).
    :param connection_timeout: Optional. Defaults to 10. The timeout in seconds to use when opening SSH connections.
    :param transfer_timeout: Optional. Defaults to 10 * 60. The timeout in seconds to use when transferring files.
    :param window_size: Optional. Set the window size on the transport used for the SFTP connection.
        Defaults to 2097152.
    :param max_packet_size: Optional. Sets the packet size on the transport used for the SFTP connection.
        Defaults to 32768.
    :param verify_host_key: Whether to verify the device host key or not when connecting.

    :raises exceptions.AuthenticationFailed: if authentication using the supplied credentials failed.
    :raises exceptions.ConnectionFailed: if connectivity to the supplied address failed.
    :raises exceptions.PrivateKeyError: if a supplied private key fails to load.

    :returns: A context manager that, when called using the `with` statement, yields a connector
        with a SFTP client.

    The returned context manager should only be called using a `with statement
    <https://docs.python.org/3.8/reference/compound_stmts.html#with>`_.
    All SFTP client operations must happen inside the `with` block. You do not need to know how to
    open or close an SFTP connection - the context manager handles all of this for you.

    You can re-use an initialized session as often as necessary. This makes it easier to
    separate offline programming logic from SFTP operations as much as possible.

    See below for a basic snippet

        .. code-block:: python

            get_connector = sftp_client.initialize(
                address="10.9.10.134",
                username="user123",
                password="mUh5f&uid@",
            )
            with get_connector() as connection:
                sftp_client.download_file(
                    connection,
                    remote_filepath="/home/ubuntu/apt-list.txt",
                    local_filepath="/tmp/apt-list.txt"
                )

    """
    return functools.partial(
        _sftp_client._connect,
        address=address,
        port=port,
        username=username,
        password=password,
        private_key=private_key,
        passphrase=passphrase,
        connection_timeout=connection_timeout,
        transfer_timeout=transfer_timeout,
        window_size=window_size,
        max_packet_size=max_packet_size,
        missing_host_key_policy=paramiko.RejectPolicy() if verify_host_key else paramiko.WarningPolicy(),
    )


@typechecked
def download_file(connection, *, source: str, destination: str):
    """ Download a file from the target device.

    :param connection: An open SFTP connection to the target device
    :param source: The file path to the file to download from the target device
    :param destination: The file path where the file should be downloaded to.

    :raises exceptions.FileNotFound: if the file is not found at the provided remote file path.
    :raises exceptions.PermissionDenied: if the user does not have permission to download the
        file to the local file path.

    See below for a basic snippet

        .. code-block:: python

            get_connector = sftp_client.initialize(
                address="10.9.10.134",
                username="user123",
                password="mUh5f&uid@",
            )
            with get_connector() as connection:
                sftp_client.download_file(
                    connection,
                    source="/home/ubuntu/apt-list.txt",
                    destination="/tmp/apt-list.txt"
                )

    .. warning::
        If the file already exists at the specified local file path, the file will be overwritten.

    """
    try:
        connection.get(source, destination)
    except FileNotFoundError:
        raise exceptions.FileNotFound(f"Unable to find file '{source}' to download.")
    except PermissionError:
        raise exceptions.PermissionDenied(
            f"Failed to download file '{source}' to '{destination}' due to insufficient permissions."
        )


@typechecked
def upload_file(connection, *, source: str, destination: str):
    """Upload a file to the target device.

    :param connection: An open SFTP connection to the target device
    :param source: The file path of the file to upload.
    :param destination: The file path where the specified file should be saved to on the target device

    :raises exceptions.FileNotFound: if the file is not found at the provided local file path.
    :raises exceptions.PermissionDenied: if the user does not have permission to upload the file to the
        remote file path.

    See below for a basic snippet

        .. code-block:: python

            get_connector = sftp_client.initialize(
                address="10.9.10.134",
                username="user123",
                password="mUh5f&uid@",
            )
            with get_connector() as connection:
                sftp_client.upload_file(
                    connection,
                    source="/home/ppa/apt-list.txt",
                    destination="/home/ubuntu/apt-list.txt"
                )

    .. warning::
        If the file already exists at the specified remote file path, the file will be overwritten.

    """
    try:
        connection.put(source, destination)
    except FileNotFoundError:
        raise exceptions.FileNotFound(f"Unable to find file '{source}' to upload to device.")
    except PermissionError:
        raise exceptions.PermissionDenied(
            f"Failed to upload file '{source}' to '{destination}' due to insufficient permissions."
        )
