import base64
import logging
import os

from collections import namedtuple
from dataclasses import dataclass
from typing import List, Dict, Union, Tuple, Optional

from .api import event, task_event
from .appliance import min_ppa_version, retired
from .exceptions import (
    EventNotConfigured,
    EventTypeDisabled,
    EventConfigurationError,
    NoMailRecipients,
    EmailEventError,
)

logger = logging.getLogger(__name__)


class EventType(namedtuple("EventType", ("protocol", "event_name"))):
    """ An event type supported by the PPA appliance.

    :param protocol: The event protocol.
    :param event_name: A friendly name for the event used in exception messages if an error occurs.
    """


class Attachment(namedtuple("Attachment", ("filename", "path"))):
    """ An attachment to be attached to an email sent via the send_email function.

    :param filename: The file name.
    :param path: The path to the file.
    """


@dataclass(frozen=True)
class Events:
    SMTP = EventType(protocol="SMTP", event_name="email")
    CEF = EventType(protocol="syslog", event_name="CEF event")
    SYSLOG = EventType(protocol="syslog", event_name="JSON Syslog event")


@dataclass(frozen=True)
class CEFEventLevels:
    INFO = "info"
    PANIC = "panic"
    FATAL = "fatal"
    ERROR = "error"
    WARNING = "warning"
    DEBUG = "debug"
    TRACE = "trace"


@min_ppa_version("2.7.0")
@task_event(
    {
        204: EventNotConfigured,
        400: EmailEventError,
        405: EventTypeDisabled,
        422: NoMailRecipients,
        500: EmailEventError,
    },
    event_type=Events.SMTP,
)
def send_email(
    *,
    subject: str,
    html: str,
    recipients: Optional[List[str]] = None,
    plain_text: Optional[str] = "Please enable HTML to view this email.",
    reason: Optional[str] = None,
    attachments: Optional[List[Attachment]] = None,
) -> None:
    return event(
        "message/smtp",
        reason,
        subject=subject,
        recipients=recipients,
        html=html,
        message=plain_text,
        files=[("file", (attachment.filename, open(attachment.path, "rb"))) for attachment in attachments]
        if attachments
        else None,
    )


@min_ppa_version("2.4.0")
@retired("2.5.0", message="Syslog changed to JSON format. Use the send_syslog_event function instead.")
@task_event({405: EventConfigurationError}, event_type=Events.CEF)
def send_cef_event(
    *,
    message: str,
    level: Optional[str] = CEFEventLevels.INFO,
    reason: Optional[str] = None,
    fields: Optional[dict] = None,
) -> None:
    return event("message/syslog", reason, message=message, level=level, fields=fields)


@min_ppa_version("2.5.0")
@task_event({405: EventConfigurationError}, event_type=Events.SYSLOG)
def send_syslog_event(
    *, message: str, level: Optional[str] = "info", reason: Optional[str] = None, details: Optional[dict] = None,
) -> None:
    # For now 'details' is sent to the appliance as 'fields' - https://github.com/Osirium/ppa/issues/445
    return event("message/syslog", reason, message=message, level=level, fields=details)


def create_wait_hook() -> Tuple[str, str]:
    secret = base64.urlsafe_b64encode(os.urandom(24)).decode("utf-8")
    return (f"{os.environ['PPA_WAIT_HOOK']}/{secret}", secret)


@task_event()
def wait(
    text: str,
    *,
    secret: str,
    element_id: Optional[str] = None,
    reason: Optional[bool] = False,
    template: Optional[str] = None,
) -> Dict[str, Union[str, list, dict]]:
    return event("hooks/wait", text, secret=secret, id=element_id, provide_reason=reason, template=template)
