from typing import TypedDict, Union, Optional, List, Dict, Any

from datacaster.classes import CastDataClass


DEFAULT_PROMPT = r"^[a-zA-Z0-9\_\-]+@[a-zA-Z0-9\_\-]+:~\$\s+$"


class SSHKeyAuth(TypedDict):
    r"""
- Contains SSH connection details for key authentication

- It is used as an _input_ to many of the actions in this plugin

- The example below shows the dictionary structure in YAML

!!! example

    ```yaml linenums="1"
    ssh_server:
      address: 1.2.3.4
      port: 22
      username: [username]
      private_key: [private_key_string]
      passphrase: [optional_passphrase]
      base_prompt: "ubuntu\-\d{3}#"
    ```

!!! tip "SSH Prompts"

    As of plugin version 1.1.0, if a __base_prompt__ expression is not supplied, PPA will use the following:

    `^[a-zA-Z0-9\_\-]+@[a-zA-Z0-9\_\-]+:~\$\s+$`

    This expression supports the default Ubuntu SSH prompt in this format:

    __ubuntu@ip-172-30-23-5:~$__

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

???+ danger "Required Keys"

    __address__: The target IP or DNS address.

    __username__: Username for authentication.

    __private_key__: RSA private key for authentication.

    __passphrase__: The passphrase for the supplied _private_key_ ___or___ _null_ if not required.


!!! tip "Optional Keys"

    __port__: The target SSH port (defaults to _22_)

    __base_prompt__: Regular expression to match the SSH prompt (defaults to the standard Ubuntu SSH prompt format)
    """

    address: str
    port: Union[str, int] = 22
    username: str
    private_key: Optional[str]
    passphrase: Optional[str]
    base_prompt: Optional[str] = DEFAULT_PROMPT


class SSHPasswordAuth(TypedDict):
    r"""
- Contains SSH connection details for password authentication

- It is used as an _input_ to many of the actions in this plugin

- The example below shows the dictionary structure in YAML

!!! example

    ```yaml linenums="1"
    ssh_server:
      address: 1.2.3.4
      port: 22
      username: [username]
      password: [password]
      base_prompt: "ubuntu\-\d{3}#"
    ```

!!! tip "SSH Prompts"

    As of plugin version 1.1.0, if a __base_prompt__ expression is not supplied, PPA will use the following:

    `^[a-zA-Z0-9\_\-]+@[a-zA-Z0-9\_\-]+:~\$\s+$`

    This expression supports the default Ubuntu SSH prompt in this format:

    __ubuntu@ip-172-30-23-5:~$__

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.


!!! danger "Required Keys"

    __address__: The target IP or DNS address.

    __username__: Username for authentication.

    __password__: Password for authentication.


!!! tip "Optional Keys"

    __port__: The target SSH port (defaults to _22_)

    __base_prompt__: Regular expression to match the SSH prompt (defaults to the standard Ubuntu SSH prompt format)

    """
    address: str
    port: Union[str, int] = 22
    username: str
    password: Optional[str]
    base_prompt: Optional[str] = DEFAULT_PROMPT


UbuntuServer = Union[SSHPasswordAuth, SSHKeyAuth]


class CommandQuestion(TypedDict):
    r"""
- Contains prompts and responses for questions asked by a command

- You can supply as many of these questions & answers as you need to

- Each question prompt ___must___ be unique per-command

- Question prompts that aren't found are ignored

__Question Handling__

The following process is started if question prompt is found.

1. The answer to the prompted question is sent
2. If the response is in `base_prompt` or _success_, processing is __complete__
3. Or if it's in another question prompt, go back to __step 1__

If a response isn't found in step 2 or 3 it will be treated as a failure.

!!! example
    {% raw %}
    ```YAML linenums="1" hl_lines="4 5 6"
    - ssh.client.run_commands:
        commands:
          - command: sudo ls --color=never -lsah
            questions:
              prompt: "\[sudo\] password for .*:"
              answer: "{{ ssh_server.password }}"
      load:
        ssh_server: ssh_server
      save: directory_listing
    ```
    {% endraw %}

???+ danger "Required Keys"

    __prompt__: A regular expression matching the question prompt

    __answer__: The command to run if the question is asked
    """

    prompt: str
    answer: str


class Command(TypedDict):
    """
- Contains command information for the [run_commands](../../ssh/ssh.client/#run_command) action

!!! example

    ```YAML linenums="1" hl_lines="2"
    - ssh.client.run_command:
        command: ls --color=never -lsah
      load:
        ssh_server: ssh_server
      save: directory_listing
    ```

???+ danger "Required Keys"

    __command__: The command to run

???+ info "Optional Keys"

    __success__: A regular expression matching the prompt when the command succeeds (defaults to the _base_prompt_)

    __questions__: A list of [CommandQuestion](#commandquestion) dictionaries (defaults to _null_)
    """

    command: str
    success: Optional[str]
    questions: Optional[List[Dict[str, Any]]]


class UserData(CastDataClass):
    username: str
    uid: int
    gid: int
    description: Optional[str] = None
    home_directory: str
    login_shell: str


class User(TypedDict):
    """
- Contains Ubuntu user account information

!!! example

    ```YAML linenums="1"
    username: john_smith
    uid: 2087
    gid: 2087
    description: John Smith
    home_directory: "/home/john_smith"
    login_shell: "/bin/bash"
    ```

!!! info "User Keys"

    __username__: username

    __uid__: user ID

    __gid__: primary group ID

    __description__: the contents of the [GECOS](https://en.wikipedia.org/wiki/Gecos_field) field

    __home_directory__: path to the user's home directory

    __login_shell__: path to the user's login shell

    """

    __annotations__ = UserData.__dict__["__annotations__"]
